import toml
from typing import Type, Any, Literal, Dict, List
from pydantic import BaseModel
from pathlib import Path
from easydict import EasyDict as edict

CfgClass = Type[Any]
PydanticCfgClass = Type[BaseModel]


def _handle_import_path(toml_file_path: Path, import_path: Path | str) -> Path:
    repls = {
        "$this_file": toml_file_path,
        "$this_dir": toml_file_path.parent,
        "$cwd": Path.cwd()
    }
    for key, value in repls.items():
        import_path = Path(str(import_path).replace(key, str(value)))
    if import_path == toml_file_path:
        raise ValueError("Import path cannot be the same as the current file")
    return import_path


def _handle_imports(imports_list: List[Path]) -> Dict[str, Any]:
    out = {}
    for import_ in imports_list:
        import_dict = _parse_config_dict(import_)
        out.update(import_dict)
    return out


def _recursive_dict_update(d: Dict[str, Any], u: Dict[str, Any]) -> Dict[str, Any]:
    for k, v in u.items():
        if isinstance(v, dict):
            d[k] = _recursive_dict_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def _handle_preamble(config_dict: Dict[str, Any], toml_file_path: Path) -> Dict[str, Any]:
    if "pre" in config_dict.keys():
        pre = config_dict["pre"]
        if "imports" in pre.keys():
            imports = pre["imports"]
            imports_ = []
            for value in imports:
                imports_.append(_handle_import_path(toml_file_path, value))
            # At this point `imports` contains actual paths
            cfg_dict_from_imports = _handle_imports(imports_)
            # Override values from imports with those from the `config_dict`:
            cfg_dict_from_imports = _recursive_dict_update(cfg_dict_from_imports, config_dict)
            return cfg_dict_from_imports
        return config_dict
    return config_dict


def _remove_preamble(config_dict: Dict[str, Any]) -> Dict[str, Any]:
    if "pre" in config_dict.keys():
        del config_dict["pre"]
    return config_dict


def _parse_config_dict(toml_file: Path):
    with open(toml_file, 'r') as file:
        config_dict = toml.load(file)
        config_dict = _handle_preamble(config_dict, toml_file)
        config_dict = _remove_preamble(config_dict)
        return config_dict


def _parse_config_kwarg_constructor(toml_file: Path, cfg_class: CfgClass):
    config_dict = _parse_config_dict(toml_file)
    config = cfg_class(**config_dict)
    return config


def _parse_config_pydantic(toml_file: Path, cfg_class: PydanticCfgClass):
    return _parse_config_kwarg_constructor(toml_file, cfg_class)


def _parse_config_easydict(toml_file: Path):
    config_dict = _parse_config_dict(toml_file)
    return edict(config_dict)


def parse_config(toml_file: Path, cfg_class: CfgClass | Literal["attr"] | None = None):
    """Takes a path object to a toml file and returns a config object.

    Args:
        toml_file (Path): path to the toml file
        cfg_class (CfgClass | Literal["attr"], optional): config loader class. Defaults to None.
            If set to `"attr"`, the config will be loaded as an `easydict` object instead
            of a conventional dictionary.

    Returns:
        An instance of the class used to load the config
    """
    match cfg_class:
        case None:
            return _parse_config_dict(toml_file)
        case "attr":
            return _parse_config_easydict(toml_file)
        case BaseModel():
            return _parse_config_pydantic(toml_file, cfg_class)
        case _:
            return _parse_config_kwarg_constructor(toml_file, cfg_class)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['confuk']

package_data = \
{'': ['*']}

install_requires = \
['easydict>=1.11,<2.0', 'pydantic>=2.5.3,<3.0.0', 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'confuk',
    'version': '0.2.0',
    'description': 'Very opinionated configuration loading package for Python projects',
    'long_description': '# Confuk\n\nThis is yet another package for managing configuration files in Python projects.\n\nAt the moment all it does is it exposes a consistent API that lets you provide a path to a TOML configuration file. It parses the config file into a dictionary by default. If a config class is provided when parsing, the class instance will be created using a dictionary of keyword arguments coming from the original TOML file.\n\nIn human words: I made this package so that I don\'t have to explicilty load, parse and return a class instance every single time I have something to do with a TOML file:\n\n```python\nfrom confuk import parse_config\nfrom pathlib import Path\nfrom somewhere import ConfigClass\n\ncfg_dict = parse_config(Path("some.toml"))  # returns a dictionary\ncfg_obj = parse_config(Path("some.toml"), ConfigClass)  # returns an instance of `ConfigClass`\n```\n\n## Installation\n\nCurrently you can build the package using Poetry:\n\n1. Clone this repo.\n2. Run `poetry build`.\n3. Grab the installable wheel from the `dist` folder and install it with `pip` or add the package as a local dependency of another project.\n\nOnce I get some time to take care of it I will add the package to PyPI so that it\'s installable via a simple `pip install confuk` command.\n',
    'author': 'Krzysztof J. Czarnecki',
    'author_email': 'kjczarne@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
